"""GWGEN: A WGEN-like weather generator for global daily weather
"""
from gwgen.version import __version__

__author__ = "Philipp Sommer and Jed Kaplan"
__maintainer__ = "Philipp Sommer"
__licence__ = "GPLv2"
__email__ = "philipp.sommer@unil.ch"
__copyright__ = "Copyright 2016, ARVE-Research"
